
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "libnm-client-impl/nm-default-libnm.h"

#include "nm-enum-types.h"

#include "nm-version-macros.h"
#include "NetworkManager.h"
#include "nm-secret-agent-old.h"
#include "nm-vpn-plugin-old.h"
GType
nm_client_instance_flags_get_type (void)
{
  static volatile gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GFlagsValue values[] = {
        { NM_CLIENT_INSTANCE_FLAGS_NONE, "NM_CLIENT_INSTANCE_FLAGS_NONE", "none" },
        { NM_CLIENT_INSTANCE_FLAGS_NO_AUTO_FETCH_PERMISSIONS, "NM_CLIENT_INSTANCE_FLAGS_NO_AUTO_FETCH_PERMISSIONS", "no-auto-fetch-permissions" },
        { NM_CLIENT_INSTANCE_FLAGS_INITIALIZED_GOOD, "NM_CLIENT_INSTANCE_FLAGS_INITIALIZED_GOOD", "initialized-good" },
        { NM_CLIENT_INSTANCE_FLAGS_INITIALIZED_BAD, "NM_CLIENT_INSTANCE_FLAGS_INITIALIZED_BAD", "initialized-bad" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_flags_register_static (g_intern_static_string ("NMClientInstanceFlags"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
nm_client_error_get_type (void)
{
  static volatile gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { NM_CLIENT_ERROR_FAILED, "NM_CLIENT_ERROR_FAILED", "failed" },
        { NM_CLIENT_ERROR_MANAGER_NOT_RUNNING, "NM_CLIENT_ERROR_MANAGER_NOT_RUNNING", "manager-not-running" },
        { NM_CLIENT_ERROR_OBJECT_CREATION_FAILED, "NM_CLIENT_ERROR_OBJECT_CREATION_FAILED", "object-creation-failed" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("NMClientError"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}

/* Generated data ends here */

